import astropy.units as u
from astropy.modeling.polynomial import Chebyshev1D
from astropy.modeling.fitting import LevMarLSQFitter

from ..fitting import fit_lines
from ..manipulation.smoothing import median_smooth
from ..spectra import SpectralRegion


__all__ = ['fit_continuum', 'fit_generic_continuum']


def fit_generic_continuum(spectrum, median_window=3, model=Chebyshev1D(3),
                          fitter=LevMarLSQFitter(),
                          exclude_regions=None, weights=None):
    """
    Basic fitting of the continuum of an input spectrum. The input
    spectrum is smoothed using a median filter to remove the spikes.

    Parameters
    ----------
    spectrum : Spectrum1D
        The spectrum object overwhich the equivalent width will be calculated.
    model : list of `~astropy.modeling.Model`
        The list of models that contain the initial guess.
    median_window : float
        The width of the median smoothing kernel used to filter the data before
        fitting the continuum. See the ``kernel_size`` parameter of
        `~scipy.signal.medfilt` for more information.
    fitter : `~astropy.fitting._FitterMeta`
        The astropy fitter to use for fitting the model.
        Default: `~astropy.modeling.fitting.LevMarLSQFitter`
    exclude_regions : list of 2-tuples
        List of regions to exclude in the fitting. Passed through
        to the fitmodels routine.
    weights : list  (NOT IMPLEMENTED YET)
        List of weights to define importance of fitting regions.

    Returns
    -------
    continuum_model
        Fitted continuum as a model of whatever class ``model`` provides.

    Notes
    -----
       * Could add functionality to set the bounds in
         ``model`` if they are not set.
       * The models in the list of ``model`` are added together and passed as a
         compound model to the `~astropy.modeling.fitting.Fitter` class instance.

    """
    # Simple median smooth to remove spikes and peaks
    spectrum_smoothed = median_smooth(spectrum, median_window)

    return fit_continuum(spectrum_smoothed, model=model, fitter=fitter,
                         exclude_regions=exclude_regions, weights=weights)


def fit_continuum(spectrum, model=Chebyshev1D(3), fitter=LevMarLSQFitter(),
                  exclude_regions=None, exclude_region_upper_bounds=False,
                  window=None, weights=None):
    """
    Entry point for fitting using the `~astropy.modeling.fitting`
    machinery.

    Parameters
    ----------
    spectrum : Spectrum1D
        The spectrum object overwhich the equivalent width will be calculated.
    model: list of `~astropy.modeling.Model`
        The list of models that contain the initial guess.
    fitter : `~astropy.fitting._FitterMeta`
        The astropy fitter to use for fitting the model.
        Default: `~astropy.modeling.fitting.LevMarLSQFitter`
    exclude_regions : list of 2-tuples
        List of regions to exclude in the fitting. Passed through
        to the fitmodels routine.
    exclude_region_upper_bounds : bool, optional
        Set to True to override the default pythonic slicing on the excluded regions (inclusive
        lower bound, exclusive upper bound) and remove spectral values less than or equal to
        the upper bound of the region rather than strictly less than the upper bound. Passed
        through to the fitmodels routine.
    window : tuple of wavelengths
        Start and end wavelengths used for fitting.
    weights : list  (NOT IMPLEMENTED YET)
        List of weights to define importance of fitting regions.

    Returns
    -------
    models : list of `~astropy.modeling.Model`
        The list of models that contain the fitted model parameters.
    """
    if weights is not None:
        raise NotImplementedError("Weights support is not yet implemented.")

    w = window
    if type(w) in [list, tuple] and all([_is_valid_sequence(x) for x in w]):
        w = SpectralRegion(w)

    # Fit the flux to the model
    continuum_spectrum = fit_lines(spectrum, model, fitter, exclude_regions,
                                   exclude_region_upper_bounds, weights, w)

    return continuum_spectrum


# Checks for sequences of of 2-tuples with Quantities
def _is_valid_sequence(value):
    if type(value) in [list, tuple]:
        return len(value) == 2 and \
               isinstance(value[0], u.Quantity) and \
               isinstance(value[1], u.Quantity)

    return False



from specutils import SpectrumList
from specutils.io.registers import data_loader

from .dc_common import FITS_FILE_EXTS, SINGLE_SPLIT_LABEL
from ..parsing_utils import read_fileobj_or_hdulist

GALAH_5EXT_CONFIG = {
    "hdus": {
        "0": {
            "purpose": "science",
            "units": {"flux_unit": "count"},
        },
        "1": {
            "purpose": "error_stdev",
            "units": {"flux_unit": "count"},
        },
        "2": {
            "purpose": "unreduced_science",
            "units": {"flux_unit": "count"},
        },
        "3": {
            "purpose": "unreduced_error_stdev",
            "units": {"flux_unit": "count"},
        },
        "4": {
            "purpose": "normalised_science",
            "units": {"flux_unit": ""},
        },
    },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": None,
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
GALAH_4EXT_CONFIG = {
    "hdus": {
        "0": {"purpose": "science"},
        "1": {"purpose": "error_stdev"},
        "2": {"purpose": "unreduced_science"},
        "3": {"purpose": "unreduced_error_stdev"},
    },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}


def identify_galah(origin, *args, **kwargs):
    """
    Identify if the current file is a GALAH file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        if "GALAH" in hdulist[0].header.get("ORIGIN"):
            return True
        return False


@data_loader(
    label="GALAH", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_galah, priority=10,
)
def galah_loader(filename):
    with read_fileobj_or_hdulist(filename) as hdulist:
        if len(hdulist) == 5:
            spectra = SpectrumList.read(
                hdulist, format=SINGLE_SPLIT_LABEL, **GALAH_5EXT_CONFIG
            )
            spectra[0].meta["galah_hdu_format"] = 5
        elif len(hdulist) == 4:
            spectra = SpectrumList.read(
                hdulist, format=SINGLE_SPLIT_LABEL, **GALAH_4EXT_CONFIG
            )
            spectra[0].meta["galah_hdu_format"] = 4
        else:
            raise RuntimeError(
                "Unknown GALAH format, has {} extensions".format(len(hdulist))
            )
        return spectra



"""
Generate small test versions of DESI files::

    python -m specutils.io.default_loaders.tests.generate_desi_test_data

The resulting files try to be self-consistent, *i.e.*, as close as possible
to a genuine DESI file with N spectra, where N is of order 10.
"""
import os
import sys
import warnings
from argparse import ArgumentParser
import numpy as np
from astropy.io import fits
from astropy.utils.data import download_file


desi_url = 'https://data.desi.lbl.gov'
downloads = {'healpix_coadd': 'public/edr/spectro/redux/fuji/healpix/sv3/dark/260/26065/coadd-sv3-dark-26065.fits',
             'healpix_spectra': 'public/edr/spectro/redux/fuji/healpix/sv3/dark/260/26065/spectra-sv3-dark-26065.fits',
             'tile_coadd': 'public/edr/spectro/redux/fuji/tiles/cumulative/169/20210419/coadd-5-169-thru20210419.fits',
             'tile_spectra': 'public/edr/spectro/redux/fuji/tiles/cumulative/169/20210419/spectra-5-169-thru20210419.fits'}


def _options():
    """Parse command-line options.
    """
    prsr = ArgumentParser(description='Generate small test versions of DESI files.')
    prsr.add_argument('-l', '--local-path', metavar='PATH', dest='path',
                      help='Find local files in PATH instead of downloading.')
    prsr.add_argument('-o', '--overwrite', action='store_true',
                      help='Overwrite any existing files.')
    prsr.add_argument('-r', '--rows', default=5, type=int, metavar='N',
                      help='Save N spectra from the original file (default %(default)s).')
    prsr.add_argument('output', metavar='PATH', help='Write files to PATH.')
    return prsr.parse_args()


def main():
    """Entry point for command-line scripts.

    Returns
    -------
    :class:`int`
        An integer suitable for passing to :func:`sys.exit`.
    """
    options = _options()
    files = list()
    for key in downloads:
        if options.path is None:
            filename = download_file(f"{desi_url}/{downloads[key]}", cache=True)
        else:
            filename = f"{options.path}/{downloads[key]}"
        if os.path.isfile(filename):
            files.append(filename)
        else:
            warnings.warn(f"Could not find {filename}!")
    for f in files:
        print(f)
        b = os.path.basename(f)
        with fits.open(f, mode='readonly') as hdulist:
            new_hdulist = hdulist.copy()
            for hdu in new_hdulist:
                try:
                    print(hdu.header['EXTNAME'])
                except KeyError:
                    print("PRIMARY")
                if hdu.header['NAXIS'] == 0:
                    pass
                elif hdu.header['EXTNAME'] == 'FIBERMAP' or hdu.header['EXTNAME'] == 'EXP_FIBERMAP' or hdu.header['EXTNAME'] == 'SCORES':
                    if hdu.header['EXTNAME'] == 'FIBERMAP':
                        targetids = hdu.data['TARGETID'][0:options.rows]
                    if hdu.header['EXTNAME'] == 'EXP_FIBERMAP':
                        exp_index = None
                        for t in targetids:
                            if exp_index is None:
                                exp_index = np.where(hdu.data['TARGETID'] == t)[0]
                            else:
                                exp_index = np.append(exp_index, np.where(hdu.data['TARGETID'] == t)[0])
                        hdu.data = hdu.data[exp_index]
                    else:
                        hdu.data = hdu.data[0:options.rows]
                    # hdu.add_checksum()
                elif hdu.header['NAXIS'] == 1:
                    pass
                elif hdu.header['NAXIS'] == 2:
                    hdu.data = hdu.data[0:options.rows, :]
                    # hdu.add_checksum()
                elif hdu.header['NAXIS'] == 3:
                    hdu.data = hdu.data[0:options.rows, :, :]
                    # hdu.add_checksum()
                if 'CHECKSUM' in hdu.header:
                    hdu.add_checksum()
            new_hdulist.writeto(os.path.join(options.output, b), output_verify='warn',
                                overwrite=options.overwrite, checksum=False)
    return 0


if __name__ == '__main__':
    sys.exit(main())



import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian1D
from astropy.tests.helper import assert_quantity_allclose

from specutils.fitting import find_lines_derivative
from specutils.spectra import Spectrum1D, SpectralRegion


def test_lower_upper():
    # Spectral region with just one range (lower and upper bound)
    sr = SpectralRegion(0.45*u.um, 0.6*u.um)

    assert sr.lower == 0.45*u.um
    assert sr.upper == 0.6*u.um

    # Spectral region with just two ranges
    sr = SpectralRegion([(0.45*u.um, 0.6*u.um), (0.8*u.um, 0.9*u.um)])

    assert sr.lower == 0.45*u.um
    assert sr.upper == 0.9*u.um

    # Spectral region with multiple ranges and not ordered
    sr = SpectralRegion([(0.3*u.um, 1.0*u.um), (0.45*u.um, 0.6*u.um),
                         (0.04*u.um, 0.05*u.um), (0.8*u.um, 0.9*u.um)])

    assert sr.lower == 0.04*u.um
    assert sr.upper == 1.0*u.um

    # Get lower bound of a single sub-region:
    assert sr[0].lower == 0.04*u.um
    assert sr[0].upper == 0.05*u.um


@pytest.mark.parametrize(
    ('center', 'width', 'lower', 'upper'),
    [(6563 * u.AA, 10 * u.AA, 6558.0 * u.AA, 6568.0 * u.AA),
     (1 * u.GHz, 0.1 * u.GHz, 1.05 * u.GHz, 0.95 * u.GHz),
     (0.5 * u.pix, 1 * u.pix, 0 * u.pix, 1 * u.pix)])
def test_from_center(center, width, lower, upper):
    # Spectral region from center with width
    sr = SpectralRegion.from_center(center=center, width=width)
    assert_quantity_allclose(sr.lower, lower)
    assert_quantity_allclose(sr.upper, upper)


@pytest.mark.parametrize(
    ('center', 'width'),
    [(6563 * u.AA, -10 * u.AA),
     (6563 * u.AA, 0 * u.AA),
     (1 * u.GHz, -0.1 * u.GHz)])
def test_from_center_error(center, width):
    with pytest.raises(ValueError):
        SpectralRegion.from_center(center=center, width=width)


def test_adding_spectral_regions():

    # Combine two Spectral regions into one:
    sr = (SpectralRegion(0.45*u.um, 0.6*u.um) +
          SpectralRegion(0.8*u.um, 0.9*u.um))

    assert set(sr.subregions) == set([(0.45*u.um, 0.6*u.um),
                                      (0.8*u.um, 0.9*u.um)])

    # In-place adding spectral regions:
    sr1 = SpectralRegion(0.45*u.um, 0.6*u.um)
    sr2 = SpectralRegion(0.8*u.um, 0.9*u.um)
    sr1 += sr2

    assert set(sr1.subregions) == set([(0.45*u.um, 0.6*u.um),
                                       (0.8*u.um, 0.9*u.um)])


def test_getitem():
    sr = SpectralRegion([(0.8*u.um, 0.9*u.um), (0.3*u.um, 1.0*u.um),
                         (0.45*u.um, 0.6*u.um), (0.04*u.um, 0.05*u.um)])

    assert sr[0].subregions == [(0.04*u.um, 0.05*u.um)]
    assert sr[1].subregions == [(0.3*u.um, 1.0*u.um)]
    assert sr[2].subregions == [(0.45*u.um, 0.6*u.um)]
    assert sr[3].subregions == [(0.8*u.um, 0.9*u.um)]
    assert sr[-1].subregions == [(0.8*u.um, 0.9*u.um)]


def test_bounds():

    # Single subregion
    sr = SpectralRegion(0.45*u.um, 0.6*u.um)
    assert sr.bounds == (0.45*u.um, 0.6*u.um)

    # Multiple subregions
    sr = SpectralRegion([(0.8*u.um, 0.9*u.um), (0.3*u.um, 1.0*u.um),
                         (0.45*u.um, 0.6*u.um), (0.04*u.um, 0.05*u.um)])
    assert sr.bounds == (0.04*u.um, 1.0*u.um)


def test_delitem():

    # Single subregion
    sr = SpectralRegion(0.45*u.um, 0.6*u.um)

    del sr[0]
    assert sr.subregions == []

    # Multiple sub-regions
    sr = SpectralRegion([(0.8*u.um, 0.9*u.um), (0.3*u.um, 1.0*u.um),
                         (0.45*u.um, 0.6*u.um), (0.04*u.um, 0.05*u.um)])
    del sr[1]

    assert sr[0].subregions == [(0.04*u.um, 0.05*u.um)]
    assert sr[1].subregions == [(0.45*u.um, 0.6*u.um)]
    assert sr[2].subregions == [(0.8*u.um, 0.9*u.um)]


def test_iterate():

    # Create the Spectral region
    subregions = [(0.8*u.um, 0.9*u.um), (0.3*u.um, 1.0*u.um),
                  (0.45*u.um, 0.6*u.um), (0.04*u.um, 0.05*u.um)]
    sr = SpectralRegion(subregions)

    # For testing, sort our subregion list.
    subregions.sort(key=lambda k: k[0])

    for ii, s in enumerate(sr):
        assert s.subregions[0] == subregions[ii]


def test_slicing():

    sr = (SpectralRegion(0.15*u.um, 0.2*u.um) +
          SpectralRegion(0.3*u.um, 0.4*u.um) +
          SpectralRegion(0.45*u.um, 0.6*u.um) +
          SpectralRegion(0.8*u.um, 0.9*u.um) +
          SpectralRegion(1.0*u.um, 1.2*u.um) +
          SpectralRegion(1.3*u.um, 1.5*u.um))

    subsr = sr[3:5]

    assert subsr[0].subregions == [(0.8*u.um, 0.9*u.um)]
    assert subsr[1].subregions == [(1.0*u.um, 1.2*u.um)]


def test_invert():
    sr = (SpectralRegion(0.15*u.um, 0.2*u.um) +
          SpectralRegion(0.3*u.um, 0.4*u.um) +
          SpectralRegion(0.45*u.um, 0.6*u.um) +
          SpectralRegion(0.8*u.um, 0.9*u.um) +
          SpectralRegion(1.0*u.um, 1.2*u.um) +
          SpectralRegion(1.3*u.um, 1.5*u.um))

    sr_inverted_expected = [(0.05*u.um, 0.15*u.um), (0.2*u.um, 0.3*u.um),
                            (0.4*u.um, 0.45*u.um), (0.6*u.um, 0.8*u.um),
                            (0.9*u.um, 1.0*u.um), (1.2*u.um, 1.3*u.um),
                            (1.5*u.um, 3.0*u.um)]

    # Invert from range.
    sr_inverted = sr.invert(0.05*u.um, 3*u.um)

    for ii, expected in enumerate(sr_inverted_expected):
        assert sr_inverted.subregions[ii] == sr_inverted_expected[ii]

    # Invert from spectrum.
    spectrum = Spectrum1D(spectral_axis=np.linspace(0.05, 3, 20)*u.um,
                          flux=np.random.random(20)*u.Jy)
    sr_inverted = sr.invert_from_spectrum(spectrum)
    for ii, expected in enumerate(sr_inverted_expected):
        assert sr_inverted.subregions[ii] == sr_inverted_expected[ii]


def test_from_list_list():
    g1 = Gaussian1D(1, 4.6, 0.2)
    g2 = Gaussian1D(2.5, 5.5, 0.1)
    g3 = Gaussian1D(-1.7, 8.2, 0.1)

    x = np.linspace(0, 10, 200)
    y = g1(x) + g2(x) + g3(x)

    spectrum = Spectrum1D(flux=y * u.Jy, spectral_axis=x * u.um)

    lines = find_lines_derivative(spectrum, flux_threshold=0.01)

    spec_reg = SpectralRegion.from_line_list(lines)
    expected = [(4.072864321608041 * u.um, 5.072864321608041 * u.um),
                (4.977386934673367 * u.um, 5.977386934673367 * u.um),
                (7.690954773869347 * u.um, 8.690954773869347 * u.um)]

    for i, reg in enumerate(expected):
        assert_quantity_allclose(reg, (spec_reg[i].lower, spec_reg[i].upper))


def test_read_write():
    sr = SpectralRegion([(0.45*u.um, 0.6*u.um), (0.8*u.um, 0.9*u.um)])
    sr.write("test_sr.ecsv")
    sr2 = SpectralRegion.read("test_sr.ecsv")
    assert list(sr2.as_table().columns) == ["lower_bound", "upper_bound"]

    sr3 = SpectralRegion.from_qtable(sr2.as_table())
    assert sr3.subregions[0] == sr.subregions[0]
    assert sr3.subregions[1] == sr.subregions[1]



# Third-party
from astropy.utils.data import download_file
from astropy.config import set_temp_cache
import pytest

# Package
from specutils.io.default_loaders.apogee import (apStar_loader, apVisit_loader,
                                                 aspcapStar_loader)


@pytest.mark.remote_data
def test_apStar_loader(tmp_path):
    apstar_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/"
                  "stars/apo25m/N7789/apStar-r12-2M00005414+5522241.fits")
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(apstar_url, cache=True)
        spectrum = apStar_loader(filename)  # noqa


@pytest.mark.remote_data
def test_apVisit_loader(tmp_path):
    apvisit_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/"
                   "visit/apo25m/N7789/5094/55874/"
                   "apVisit-r12-5094-55874-123.fits")
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(apvisit_url, cache=True)
        spectrum = apVisit_loader(filename)  # noqa


@pytest.mark.remote_data
def test_aspcapStar_loader(tmp_path):
    aspcap_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/aspcap/r12/"
                  "l33/apo25m/N7789/aspcapStar-r12-2M00005414+5522241.fits")
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(aspcap_url, cache=True)
        spectrum = aspcapStar_loader(filename)  # noqa



import copy

import numpy as np
from astropy import units as u
from astropy.modeling.models import Shift
from astropy.modeling.tabular import Tabular1D
from gwcs import WCS as GWCS
from gwcs import coordinate_frames as cf


class SpectralGWCS(GWCS):
    """
    This is a placeholder lookup-table GWCS created when a :class:`~specutils.Spectrum1D` is
    instantiated with a ``spectral_axis`` and no WCS.
    """
    def __init__(self, *args, **kwargs):
        self.original_unit = kwargs.pop("original_unit", "")
        super().__init__(*args, **kwargs)

    def copy(self):
        """
        Return a shallow copy of the object.

        Convenience method so user doesn't have to import the
        :mod:`copy` stdlib module.

        .. warning::
            Use `deepcopy` instead of `copy` unless you know why you need a
            shallow copy.
        """
        return copy.copy(self)

    def deepcopy(self):
        """
        Return a deep copy of the object.

        Convenience method so user doesn't have to import the
        :mod:`copy` stdlib module.
        """
        return copy.deepcopy(self)

    def pixel_to_world(self, *args, **kwargs):
        if self.original_unit == '':
            return u.Quantity(super().pixel_to_world_values(*args, **kwargs))
        return super().pixel_to_world(*args, **kwargs).to(
            self.original_unit, equivalencies=u.spectral())


def refraction_index(wavelength, method='Morton2000', co2=None):
    """
    Calculates the index of refraction of dry air at standard temperature
    and pressure, at different wavelengths, using different methods.

    Parameters
    ----------
    wavelength : `Quantity` object (number or sequence)
        Vacuum wavelengths with an astropy.unit.

    method : str, optional
        Method used to convert wavelengths. Options are:

        'Morton2000' (default) - from Morton (2000, ApJS, 130, 403), eqn 8. Used by VALD,
            the Vienna Atomic Line Database. Very similar to Edlen (1966).

        'Griesen2006' - from Greisen et al. (2006, A&A 446, 747),
            eqn. 65, standard used by International Union of Geodesy and Geophysics

        'Edlen1953' - from Edlen (1953, J. Opt. Soc. Am, 43, 339). Standard
            adopted by IAU (resolution No. C15, Commission 44, XXI GA, 1991),
            which refers to Oosterhoff (1957) that uses Edlen (1953). Also used
            by Morton (1991, ApJS, 77, 119), which is frequently cited as IAU source.

        'Edlen1966' - from Edlen (1966, Metrologia 2, 71), rederived constants
            from optical and near UV data.

        'PeckReeder1972' - from Peck & Reeder (1972, J. Opt. Soc. 62), derived
            from additional infrared measurements (up to 1700 nm).

        'Ciddor1996' - from Ciddor (1996, Appl. Opt. 35, 1566). Based on
            Peck & Reeder (1972), but updated to account for the changes in
            the international temperature scale and adjust the results for
            CO2 concentration. Arguably most accurate conversion available.

        Note that all options except for 'Griesen2006' have singularities in the far
        UV. 'Griesen2006' gives values that are slightly inconsistent with the
        other methods (~0.07 Angstrom difference at visible wavelengths), but it is
        the best option in the FUV due to the mathematical singularities in the others.
        See https://specutils.readthedocs.io/en/latest/wcs_utils.html for more detail, or
        https://github.com/astropy/specutils/issues/1162 for additional context.

    co2 : number, optional
        CO2 concentration in ppm. Only used for method='Ciddor1996'. If not
        given, a default concentration of 450 ppm is used.

    Returns
    -------
    refr : number or sequence
        Index of refraction at each given air wavelength.
    """
    VALID_METHODS = ['Griesen2006', 'Edlen1953', 'Edlen1966', 'Morton2000',
                     'PeckReeder1972', 'Ciddor1996']
    assert isinstance(method, str), 'method must be a string'
    if method != 'Griesen2006' and wavelength.min() < 200 * u.nm:
        raise ValueError("The chosen method is invalid for wavelengths below 250 nm."
                         " 'Griesen2006' is the only option for this wavelength range -"
                         " see the specutils.utils.wcs_utils.refraction_index docstring"
                         " for more detail.")
    method = method.lower()
    sigma2 = (1 / wavelength.to(u.um).value)**2
    if method == 'griesen2006':
        refr = 1e-6 * (287.6155 + 1.62887 * sigma2 + 0.01360 * sigma2**2)
    elif method == 'edlen1953':
        refr = 6.4328e-5 + 2.94981e-2 / (146 - sigma2) + 2.5540e-4 / (41 - sigma2)
    elif method == 'edlen1966':
        refr = 8.34213e-5 + 2.406030e-2 / (130 - sigma2) + 1.5997e-4 / (38.9 - sigma2)
    elif method == 'morton2000':
        refr = 8.34254e-5 + 2.406147e-2 / (130 - sigma2) + 1.5998e-4 / (38.9 - sigma2)
    elif method == 'peckreeder1972':
        refr = 5.791817e-2 / (238.0185 - sigma2) + 1.67909e-3 / (57.362 - sigma2)
    elif method == 'ciddor1996':
        refr = 5.792105e-2 / (238.0185 - sigma2) + 1.67917e-3 / (57.362 - sigma2)
        if co2:
            refr *= 1 + 0.534e-6 * (co2 - 450)
    else:
        raise ValueError("Method must be one of " + ", ".join(VALID_METHODS))
    return refr + 1


def vac_to_air(wavelength, method='Morton2000', co2=None):
    """
    Converts vacuum to air wavelengths using different methods.

    Parameters
    ----------
    wavelength : `Quantity` object (number or sequence)
        Vacuum wavelengths with an astropy.unit.
    method : str, optional
        One of the methods in refraction_index(), default is 'Morton2000'.
    co2 : number, optional
        Atmospheric CO2 concentration in ppm. Only used for method='Ciddor1996'.
        If not given, a default concentration of 450 ppm is used.

    Returns
    -------
    air_wavelength : `Quantity` object (number or sequence)
        Air wavelengths with the same unit as wavelength.
    """
    refr = refraction_index(wavelength, method=method, co2=co2)
    return wavelength / refr


def air_to_vac(wavelength, scheme='inversion', method='Morton2000', co2=None,
               precision=1e-12, maxiter=30):
    """
    Converts air to vacuum wavelengths using different methods.

    Parameters
    ----------
    wavelength : `Quantity` object (number or sequence)
        Air wavelengths with an astropy.unit.

    scheme : str, optional
        How to convert from vacuum to air wavelengths. Options are:

            * 'inversion' (default) - result is simply the inversion (1 / n) of the
              refraction index of air. Griesen et al. (2006) report that the error
              in naively inverting is less than 10^-9.

            * 'Piskunov' - uses an analytical solution derived by Nikolai Piskunov
              and used by the Vienna Atomic Line Database (VALD).

            * 'iteration' - uses an iterative scheme to invert the index of refraction.

    method : str, optional
        Only used if scheme is 'inversion' or 'iteration'. One of the methods
        in `~specutils.utils.wcs_utils.refraction_index`, default is 'Morton2000'

    co2 : number, optional
        Atmospheric CO2 concentration in ppm. Only used if scheme='inversion' and
        method='Ciddor1996'. If not given, a default concentration of 450 ppm is used.

    precision : float
        Maximum fractional value in refraction conversion beyond at which iteration will
        be stopped. Only used if scheme='iteration'.

    maxiter : integer
        Maximum number of iterations to run. Only used if scheme='iteration'.

    Returns
    -------
    vac_wavelength : `Quantity` object (number or sequence)
        Vacuum wavelengths with the same unit as wavelength.
    """
    VALID_SCHEMES = ['inversion', 'iteration', 'piskunov']
    assert isinstance(scheme, str), 'scheme must be a string'
    scheme = scheme.lower()
    if scheme == 'inversion':
        refr = refraction_index(wavelength, method=method, co2=co2)
    elif scheme == 'piskunov':
        wlum = wavelength.to(u.angstrom).value
        sigma2 = (1e4 / wlum)**2
        refr = (8.336624212083e-5 + 2.408926869968e-2 / (130.1065924522 - sigma2) +
                1.599740894897e-4 / (38.92568793293 - sigma2)) + 1
    elif scheme == 'iteration':
        # Refraction index is a function of vacuum wavelengths.
        # Iterate to get index of refraction that gives air wavelength that
        # is consistent with the reverse transformation.
        counter = 0
        result = wavelength.copy()
        refr = refraction_index(wavelength, method=method, co2=co2)
        while True:
            counter += 1
            diff = wavelength * refr - result
            if abs(diff.max().value) < precision:
                break
            if counter > maxiter:
                raise RuntimeError("Reached maximum number of iterations "
                                   "without reaching desired precision level.")
            result += diff
            refr = refraction_index(result, method=method, co2=co2)
    else:
        raise ValueError("Method must be one of " + ", ".join(VALID_SCHEMES))
    return wavelength * refr


def air_to_vac_deriv(wavelength, method='Griesen2006'):
    """
    Calculates the derivative d(wave_vacuum) / d(wave_air) using different
    methods.

    Parameters
    ----------
    wavelength : `Quantity` object (number or sequence)
        Air wavelengths with an astropy.unit.

    method : str, optional
        Method used to convert wavelength derivative. Options are:
        'Griesen2006' (default) - from Greisen et al. (2006, A&A 446, 747), eqn. 66.

    Returns
    -------
    wave_deriv : `Quantity` object (number or sequence)
        Derivative d(wave_vacuum) / d(wave_air).
    """
    assert method.lower() == 'griesen2006', "Only supported method is 'Griesen2006'"
    wlum = wavelength.to(u.um).value
    return (1 + 1e-6 * (287.6155 - 1.62887 / wlum**2 - 0.04080 / wlum**4))


def gwcs_from_array(array):
    """
    Create a new WCS from provided tabular data. This defaults to being
    a GWCS object.
    """
    orig_array = u.Quantity(array)

    coord_frame = cf.CoordinateFrame(naxes=1,
                                     axes_type=('SPECTRAL',),
                                     axes_order=(0,))
    spec_frame = cf.SpectralFrame(unit=array.unit, axes_order=(0,))

    # In order for the world_to_pixel transformation to automatically convert
    # input units, the equivalencies in the look up table have to be extended
    # with spectral unit information.
    SpectralTabular1D = type("SpectralTabular1D", (Tabular1D,),
                             {'input_units_equivalencies': {'x0': u.spectral()}})

    forward_transform = SpectralTabular1D(np.arange(len(array)),
                                          lookup_table=array)
    # If our spectral axis is in descending order, we have to flip the lookup
    # table to be ascending in order for world_to_pixel to work.
    if len(array) == 0 or array[-1] > array[0]:
        forward_transform.inverse = SpectralTabular1D(
            array, lookup_table=np.arange(len(array)))
    else:
        forward_transform.inverse = SpectralTabular1D(
                array[::-1], lookup_table=np.arange(len(array))[::-1])

    tabular_gwcs = SpectralGWCS(original_unit = orig_array.unit,
                                forward_transform=forward_transform,
                                input_frame=coord_frame,
                                output_frame=spec_frame)

    # Store the intended unit from the origin input array
    #     tabular_gwcs._input_unit = orig_array.unit

    return tabular_gwcs


def gwcs_slice(self, item):
    """
    This is a bit of a hack in order to fix the slicing of the WCS
    in the spectral dispersion direction.  The NDData slices properly
    but the spectral dispersion result was not.

    There is code slightly downstream that sets the *number* of entries
    in the dispersion axis, this is just needed to shift to the correct
    starting element.

    When WCS gets the ability to do slicing then we might be able to
    remove this code.
    """
    # Create shift of x-axis
    if isinstance(item, int):
        shift = item
    elif isinstance(item, slice):
        shift = item.start
    else:
        raise TypeError('Unknown index type {}, must be int or slice.'.format(item))

    # Create copy as we need to modify this and return it.
    new_wcs = copy.deepcopy(self)

    if shift == 0:
        return new_wcs

    shifter = Shift(shift)

    # Get the current forward transform
    forward = new_wcs.forward_transform

    # Set the new transform
    new_wcs.set_transform(new_wcs.input_frame,
                          new_wcs.output_frame,
                          shifter | forward)

    return new_wcs



import numpy as np
from astropy import units as u

__all__ = ['QuantityModel']
_NUMPY_COPY_IF_NEEDED = False if np.__version__.startswith("1.") else None

_NUMPY_COPY_IF_NEEDED = False if np.__version__.startswith("1.") else None


class QuantityModel:
    """
    The QuantityModel was created to wrap `~astropy.modeling.models` that do
    not have the ability to use `~astropy.units` in the parameters.

    Parameters
    ----------
    unitless_model : `~astropy.modeling.Model`
        A model that does not have units

    input_units : `~astropy.units`
        Units for the dispersion axis

    return_units : `~astropy.units`
        Units for the flux axis

    Notes
    -----
    When Astropy's modeling is updated so *all* models have the ability
    to have `~astropy.units.Quantity` on all parameters, then this will
    not be needed.
    """
    def __init__(self, unitless_model, input_units, return_units):
        # should  check that it's unitless somehow!
        self.unitless_model = unitless_model

        # we use the dict because now this "shadows" the unitless model's
        # input_units/ return_units
        self.__dict__['input_units'] = input_units
        self.__dict__['return_units'] = return_units

    def __hasattr_(self, nm):
        if nm in self.__dict__ or hasattr(self, self.unitless_model):
            return True

        return False

    def __getattr__(self, nm):
        if nm != 'unitless_model' and hasattr(self.unitless_model, nm):
            return getattr(self.unitless_model, nm)
        else:
            raise AttributeError("'{}' object has no attribute '{}'"
                                 "".format(self.__class__.__name__, nm))

    def __setattr__(self, nm, val):
        if nm != 'unitless_model' and hasattr(self.unitless_model, nm):
            setattr(self.unitless_model, nm, val)
        else:
            super().__setattr__(nm, val)

    def __delattr__(self, nm):
        if hasattr(self.unitless_model, nm):
            delattr(self.unitless_model, nm)
        else:
            super().__delattr__(nm)

    def __dir__(self):
        thisdir = super().__dir__()
        modeldir = dir(self.unitless_model)
        return sorted(list(thisdir) + list(modeldir))

    def __repr__(self):
        return ('<QuantityModel {}, input_units={}, '
                'return_units={}>'.format(repr(self.unitless_model)[1:-1],
                                          self.input_units, self.return_units))

    def __call__(self, x, *args, **kwargs):
        unitlessx = x.to(self.input_units).value
        result = self.unitless_model(unitlessx, *args, **kwargs)
        return u.Quantity(result, self.return_units, copy=_NUMPY_COPY_IF_NEEDED)



import numpy as np
import warnings

from astropy.utils.exceptions import AstropyUserWarning
from astropy.coordinates import SpectralCoord
from ..spectra import Spectrum1D, SpectralRegion

__all__ = ['excise_regions', 'linear_exciser', 'spectrum_from_model']


def true_exciser(spectrum, region, inclusive_upper=False):
    """
    Basic spectral excise method where the array elements in the spectral
    region defined by the parameter ``region`` (a `~specutils.SpectralRegion`)
    will be deleted from all applicable elements of the Spectrum1D object:
    flux, spectral_axis, mask, and uncertainty. If multiple subregions are
    defined in ``region``, all the subregions will be excised.

    Other methods could be defined by the user to do other types of excision.

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        The `~specutils.Spectrum1D` object to which the excision will be applied.

    region : `~specutils.SpectralRegion`
        The region of the spectrum to remove.

    inclusive_upper : bool, optional
        Set to True to override the default pythonic slicing on the regions (inclusive
        lower bound, exclusive upper bound) to include the specified upper bound instead.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Output `~specutils.Spectrum1D` with the region excised.

    Raises
    ------
    ValueError
       In the case that ``spectrum`` and ``region`` are not the correct types.

    """

    spectral_axis = spectrum.spectral_axis
    excise_indices = None

    for subregion in region:
        # Find the indices of the spectral_axis array corresponding to the subregion
        if inclusive_upper:
            region_mask = (spectral_axis >= subregion.lower) & (spectral_axis <= subregion.upper)
        else:
            region_mask = (spectral_axis >= subregion.lower) & (spectral_axis < subregion.upper)

        temp_indices = np.nonzero(region_mask)[0]
        if excise_indices is None:
            excise_indices = temp_indices
        else:
            excise_indices = np.hstack(excise_indices, temp_indices)

    new_flux = np.delete(spectrum.flux, excise_indices)
    new_spectral_axis = np.delete(spectrum.spectral_axis, excise_indices)

    if spectrum.mask is not None:
        new_mask = np.delete(spectrum.mask, excise_indices)
    else:
        new_mask = None

    if spectrum.uncertainty is not None:
        new_uncertainty = spectrum.uncertainty.__class__(
            np.delete(spectrum.uncertainty.array, excise_indices),
            unit=spectrum.uncertainty.unit)
    else:
        new_uncertainty = None

    # Return a new object with the regions excised.
    return Spectrum1D(flux=new_flux,
                      spectral_axis=new_spectral_axis,
                      uncertainty=new_uncertainty,
                      mask=new_mask,
                      wcs=spectrum.wcs,
                      velocity_convention=spectrum.velocity_convention,
                      rest_value=spectrum.rest_value if not isinstance(new_spectral_axis, SpectralCoord) else None,
                      radial_velocity=spectrum.radial_velocity if not isinstance(new_spectral_axis, SpectralCoord) else None)


def linear_exciser(spectrum, region, inclusive_upper=False):
    """
    Basic spectral excise method where the spectral region defined by the
    parameter ``region`` (a `~specutils.SpectralRegion`) will result
    in the flux between those regions set to a linear ramp of the
    two points immediately before and after the start and end of the region.

    Other methods could be defined by the user to do other types of excision.

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        The `~specutils.Spectrum1D` object to which the excision will be applied.

    region : `~specutils.SpectralRegion`
        The region of the spectrum to replace.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Output `~specutils.Spectrum1D` with the region excised.

    Raises
    ------
    ValueError
       In the case that ``spectrum`` and ``region`` are not the correct types.

    """

    spectral_axis = spectrum.spectral_axis.copy()
    flux = spectrum.flux.copy()
    modified_flux = flux
    if spectrum.uncertainty is not None:
        new_uncertainty = spectrum.uncertainty.copy()
    else:
        new_uncertainty = None

    # Need to add a check that the subregions don't overlap, since that could
    # cause undesired results. For now warn if there is more than one subregion
    if len(region) > 1:
        # Raise a warning if the SpectralRegion has more than one subregion, since
        # the handling for this is perhaps unexpected
        warnings.warn("A SpectralRegion with multiple subregions was provided as "
             "input. This may lead to undesired behavior with linear_exciser if "
             "the subregions overlap.",
             AstropyUserWarning)

    for subregion in region:
        # Find the indices of the spectral_axis array corresponding to the subregion
        if inclusive_upper:
            region_mask = (spectral_axis >= subregion.lower) & (spectral_axis <= subregion.upper)
        else:
            region_mask = (spectral_axis >= subregion.lower) & (spectral_axis < subregion.upper)

        inclusive_indices = np.nonzero(region_mask)[0]
        # Now set the flux values for these indices to be a
        # linear range
        s, e = max(inclusive_indices[0]-1, 0), min(inclusive_indices[-1]+1,
                                               spectral_axis.size-1)

        modified_flux[s:e+1] = np.linspace(flux[s], flux[e], modified_flux[s:e+1].size)

        # Add the uncertainty of the two linear interpolation endpoints in
        # quadrature and apply to the excised region.
        if new_uncertainty is not None:
            new_uncertainty[s:e] = np.sqrt(spectrum.uncertainty[s]**2 + spectrum.uncertainty[e]**2)

    # Return a new object with the regions excised.
    return Spectrum1D(flux=modified_flux,
                      spectral_axis=spectral_axis,
                      uncertainty=new_uncertainty,
                      wcs=spectrum.wcs,
                      mask=spectrum.mask,
                      velocity_convention=spectrum.velocity_convention,
                      rest_value=spectrum.rest_value if not isinstance(spectral_axis, SpectralCoord) else None,
                      radial_velocity=spectrum.radial_velocity if not isinstance(spectral_axis, SpectralCoord) else None)


def excise_regions(spectrum, regions, exciser=true_exciser, inclusive_upper=False):
    """
    Method to remove or replace the flux in the defined regions of the spectrum
    depending on the function provided in the ``exciser`` argument.

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        The `~specutils.Spectrum1D` object to which the excision will be applied.

    regions : list of `~specutils.SpectralRegion`
        Each element of the list is a `~specutils.SpectralRegion`. The flux
        between the lower and upper spectral axis value of each region will be "cut out"
        and optionally replaced with interpolated values using the ``exciser`` method.
        Note that non-overlapping regions should be provided as separate
        `~specutils.SpectralRegion` objects in this list, not as sub-regions
        in a single object in the list.

    exciser : function
        Method that takes the spectrum and region and does the excising. Other
        methods could be defined and used by this routine.
        default: true_exciser

    inclusive_upper : bool, optional
        Set to True to override the default pythonic slicing on the regions (inclusive
        lower bound, exclusive upper bound) to include the specified upper bound instead.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Output `~specutils.Spectrum1D` which has the regions excised.

    Raises
    ------
    ValueError
       In the case that ``spectrum`` and ``regions`` are not the correct types.

    """

    # Parameter checks
    if not isinstance(spectrum, Spectrum1D):
        raise ValueError('The spectrum parameter must be Spectrum1D object.')

    for region in regions:
        spectrum = excise_region(spectrum, region, exciser, inclusive_upper=inclusive_upper)

    return spectrum


def excise_region(spectrum, region, exciser=true_exciser, inclusive_upper=False):
    """
    Method to remove or replace the flux in the defined regions of the spectrum
    depending on the function provided in the ``exciser`` argument.

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        The `~specutils.Spectrum1D` object to which the smoothing will be applied.

    region : `~specutils.SpectralRegion`
        A `~specutils.SpectralRegion` object defining the region to excise.
        If excising multiple regions is desired, they should be input as a
        list of separate `~specutils.SpectralRegion` objects to
        ``excise_regions``, not as subregions defined in a single
        `~specutils.SpectralRegion`.

    exciser: method
        Method that takes the spectrum and region and does the excising. Other
        methods could be defined and used by this routine.
        default: true_exciser

    inclusive_upper : bool, optional
        Set to True to override the default pythonic slicing on the regions (inclusive
        lower bound, exclusive upper bound) to include the specified upper bound instead.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Output `~specutils.Spectrum1D` with the region excised.

    Raises
    ------
    ValueError
       In the case that ``spectrum`` and ``region`` are not the correct types.

    """

    # Parameter checks
    if not isinstance(spectrum, Spectrum1D):
        raise ValueError('The spectrum parameter must be a Spectrum1D object.')

    if not isinstance(region, SpectralRegion):
        raise ValueError('The region parameter must be a SpectralRegion object.')

    #
    #  Call the exciser method
    #

    return exciser(spectrum, region, inclusive_upper=inclusive_upper)


def spectrum_from_model(model_input, spectrum):
    """
    This method will create a `~specutils.Spectrum1D` object
    with the flux defined by calling the input ``model``. All
    other parameters for the output `~specutils.Spectrum1D` object
    will be the same as the input `~specutils.Spectrum1D` object.

    Parameters
    ----------
    model : `~astropy.modeling.Model`
        The input model or compound model from which flux is calculated.

    spectrum : `~specutils.Spectrum1D`
        The `~specutils.Spectrum1D` object to use as the model template.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        Output `~specutils.Spectrum1D` which is copy of the one passed in with the updated flux.
        The uncertainty will not be copied as it is not necessarily the same.

    """

    # If the input model has units then we will call it normally.
    if getattr(model_input, model_input.param_names[0]).unit is not None:
        flux = model_input(spectrum.spectral_axis)

    # If the input model does not have units, then assume it is in
    # the same units as the input spectrum.
    else:
        flux = model_input(spectrum.spectral_axis.value)*spectrum.flux.unit

    return Spectrum1D(flux=flux,
                      spectral_axis=spectrum.spectral_axis,
                      wcs=spectrum.wcs,
                      velocity_convention=spectrum.velocity_convention,
                      rest_value=spectrum.rest_value)



"""Register reader functions for various spectral formats."""
from typing import Optional

import numpy as np
from astropy.units import Unit, Quantity, Angstrom
from astropy.nddata import StdDevUncertainty, InverseVariance
from astropy.io.fits import HDUList, BinTableHDU, ImageHDU

from ...spectra import Spectrum1D, SpectrumList
from ..registers import data_loader
from ..parsing_utils import read_fileobj_or_hdulist

__all__ = [
    "load_sdss_apVisit_1D",
    "load_sdss_apVisit_list",
    "load_sdss_apStar_1D",
    "load_sdss_apStar_list",
    "load_sdss_spec_1D",
    "load_sdss_spec_list",
    "load_sdss_mwm_1d",
    "load_sdss_mwm_list",
]


def apVisit_identify(origin, *args, **kwargs):
    """
    Check whether given file is FITS. This is used for Astropy I/O
    Registry.

    Updated for SDSS-V outputs.
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        # Test if fits has extension of type BinTable and check for
        # apVisit-specific keys
        return (hdulist[0].header.get("SURVEY").lower() == "sdss-v"
                and len(hdulist) > 4
                and hdulist[1].header.get("BUNIT", "none").startswith("Flux")
                and hdulist[2].header.get("BUNIT", "none").startswith("Flux")
                and hdulist[4].header.get("BUNIT", "none").startswith("Wave"))


def apStar_identify(origin, *args, **kwargs):
    """
    Check whether given file is FITS. This is used for Astropy I/O
    Registry.

    Updated for SDSS-V outputs.
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        # Test if fits has extension of type BinTable and check for
        # apogee-specific keys + keys for individual apVisits
        return ("APRED" in hdulist[0].header.keys() and len(hdulist) > 9
                and "NVISITS" in hdulist[0].header.keys()
                and isinstance(hdulist[1], ImageHDU)
                and hdulist[1].header.get("BUNIT", "none").startswith("Flux")
                and hdulist[2].header.get("BUNIT", "none").startswith("Err"))


def spec_sdss5_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS and has SDSS-V spec type
    BINTABLE in first extension. This is used for Astropy I/O Registry.

    Derived from SDSS-III/IV method in sdss.py. Single change (FIBERID key).
    """
    # Test if fits has extension of type BinTable and check for spec-specific keys
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (
            (hdulist[0].header.get("TELESCOP").lower() == "sdss 2.5-m")  # and
            # hdulist[0].header.get("OBSERVAT").lower() in ["apo", "lco"]
            and (hdulist[1].header.get("TTYPE1").lower() == "flux") and
            (hdulist[1].header.get("TTYPE2").lower() == "loglam") and
            (len(hdulist) > 1) and (isinstance(hdulist[1], BinTableHDU)) and
            (hdulist[1].header.get("TTYPE3").lower() == "ivar"))


def mwm_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS and has SDSS-V mwm type
    BINTABLE in all 4 extensions. This is used for Astropy I/O Registry.
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (("V_ASTRA" in hdulist[0].header.keys()) and len(hdulist) > 0
                and ("SDSS_ID" in hdulist[0].header.keys())
                and (isinstance(hdulist[i], BinTableHDU) for i in range(1, 5)))


def _wcs_log_linear(naxis, cdelt, crval):
    """
    Convert WCS from log to linear.
    """
    return 10**(np.arange(naxis) * cdelt + crval)


def _fetch_flux_unit(hdu):
    """
    Fetch the flux unit by accessing the BUNIT key of the given HDU.

    Parameters
    ----------
    hdu : HDUImage
        Flux or e_flux HDUImage object.

    Returns
    -------
    flux_unit : Unit
        Flux unit as Astropy Unit object.

    """
    flux_unit = (hdu.header.get("BUNIT").replace("(", "").replace(
        ")", "").split(" ")[-2:])
    flux_unit = "".join(flux_unit)
    if "Ang" in flux_unit and "strom" not in flux_unit:
        flux_unit = flux_unit.replace("Ang", "Angstrom")

    flux_unit = flux_unit.replace("s/cm^2/Angstrom", "(s cm2 Angstrom)")

    return Unit(flux_unit)


# APOGEE files
@data_loader(
    "SDSS-V apStar",
    identifier=apStar_identify,
    dtype=Spectrum1D,
    priority=10,
    extensions=["fits"],
)
def load_sdss_apStar_1D(file_obj, idx: int = 0, **kwargs):
    """
    Load an apStar file as a Spectrum1D.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList.
    idx : int
        The specified visit to load. Defaults to coadd (index 0).

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file

    """
    # intialize slicer

    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # Obtain spectral axis from header (HDU[1])
        wavelength = _wcs_log_linear(
            hdulist[1].header.get("NAXIS1"),
            hdulist[1].header.get("CDELT1"),
            hdulist[1].header.get("CRVAL1"),
        )
        spectral_axis = Quantity(wavelength, unit=Angstrom)

        # Obtain flux and e_flux from data (HDU[1] & HDU[2])
        flux_unit = _fetch_flux_unit(hdulist[1])
        flux = Quantity(hdulist[1].data[idx], unit=flux_unit)
        e_flux = StdDevUncertainty(hdulist[2].data[idx])

        # reduce flux array if 1D in 2D np array
        # NOTE: bypasses jdaviz specviz NotImplementedError, but could be the expected output for specutils
        if flux.shape[0] == 1:
            flux = flux[0]
            e_flux = e_flux[0]

        # Add header + bitmask
        meta = dict()
        meta["header"] = hdulist[0].header
        # SDSS masks are arrays of bit values storing multiple bool conds
        mask = hdulist[3].data[idx]
        # NOTE: specutils considers 0/False as valid values, simlar to numpy convention
        mask = mask != 0

    return Spectrum1D(
        spectral_axis=spectral_axis,
        flux=flux,
        uncertainty=e_flux,
        mask=mask,
        meta=meta,
    )


@data_loader(
    "SDSS-V apStar multi",
    identifier=apStar_identify,
    dtype=SpectrumList,
    priority=10,
    extensions=["fits"],
)
def load_sdss_apStar_list(file_obj, **kwargs):
    """
    Load an apStar file as a SpectrumList.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList.

    Returns
    -------
    SpectrumList
        The spectra contained in the file
    """
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        nvisits = hdulist[0].header.get("NVISITS")
        if nvisits <= 1:
            raise ValueError(
                "Only 1 visit in this file. Use Spectrum1D.read() instead.")
        return SpectrumList([
            load_sdss_apStar_1D(file_obj, idx=i, **kwargs)
            for i in range(nvisits)
        ])


@data_loader(
    "SDSS-V apVisit",
    identifier=apVisit_identify,
    dtype=Spectrum1D,
    priority=10,
    extensions=["fits"],
)
def load_sdss_apVisit_1D(file_obj, **kwargs):
    """
    Load an apVisit file as a Spectrum1D.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList.

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file
    """
    flux_unit = Unit("1e-17 erg / (Angstrom cm2 s)")

    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # Fetch and flatten spectrum parameters
        spectral_axis = Quantity(hdulist[4].data.flatten(), unit=Angstrom)
        flux_unit = _fetch_flux_unit(hdulist[1])
        flux = Quantity(hdulist[1].data.flatten(), unit=flux_unit)
        e_flux = StdDevUncertainty(hdulist[2].data.flatten())

        # Get metadata and attach bitmask and MJD
        meta = dict()
        meta["header"] = hdulist[0].header
        mask = hdulist[3].data.flatten()
        # NOTE: specutils considers 0/False as valid values, simlar to numpy convention
        mask = mask != 0

    return Spectrum1D(spectral_axis=spectral_axis,
                      flux=flux,
                      mask=mask,
                      uncertainty=e_flux,
                      meta=meta)


@data_loader(
    "SDSS-V apVisit multi",
    identifier=apVisit_identify,
    dtype=SpectrumList,
    priority=10,
    extensions=["fits"],
)
def load_sdss_apVisit_list(file_obj, **kwargs):
    """
    Load an apVisit file as a SpectrumList.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList.

    Returns
    -------
    SpectrumList
        The spectra from each chip contained in the file
    """
    spectra = SpectrumList()
    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # Get metadata
        flux_unit = _fetch_flux_unit(hdulist[1])
        common_meta = dict()
        common_meta["header"] = hdulist[0].header

        for chip in range(hdulist[1].data.shape[0]):
            # Fetch spectral axis and flux, and E_flux
            spectral_axis = Quantity(hdulist[4].data[chip], unit=Angstrom)
            flux = Quantity(hdulist[1].data[chip], unit=flux_unit)
            e_flux = StdDevUncertainty(hdulist[2].data[chip])

            # Copy metadata for each, adding chip bitmask and MJD to each
            meta = common_meta.copy()
            mask = hdulist[3].data[chip]
            # NOTE: specutils considers 0/False as valid values, simlar to numpy convention
            mask = mask != 0

            spectra.append(
                Spectrum1D(
                    spectral_axis=spectral_axis,
                    flux=flux,
                    mask=mask,
                    uncertainty=e_flux,
                    meta=meta,
                ))

    return spectra


# BOSS REDUX products (specLite, specFull, custom coadd files, etc)

@data_loader(
    "SDSS-V spec",
    identifier=spec_sdss5_identify,
    dtype=Spectrum1D,
    priority=5,
    extensions=["fits"],
)
def load_sdss_spec_1D(file_obj, *args, hdu: Optional[int] = None, **kwargs):
    """
    Load a given BOSS spec file as a Spectrum1D object.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList..
    hdu : int
        The specified HDU to load a given spectra from.

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file at the specified HDU.
    """
    if hdu is None:
        # TODO: how should we handle this -- multiple things in file, but the user cannot choose.
        hdu = 1  # defaulting to coadd
        # raise ValueError("HDU not specified! Please specify a HDU to load.")
    elif hdu in [2, 3, 4]:
        raise ValueError("Invalid HDU! HDU{} is not spectra.".format(hdu))
    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # directly load the coadd at HDU1
        return _load_BOSS_HDU(hdulist, hdu, **kwargs)


@data_loader(
    "SDSS-V spec multi",
    identifier=spec_sdss5_identify,
    dtype=SpectrumList,
    priority=5,
    extensions=["fits"],
)
def load_sdss_spec_list(file_obj, **kwargs):
    """
    Load a given BOSS spec file as a SpectrumList object.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList..

    Returns
    -------
    SpectrumList
        The spectra contained in the file.
    """
    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        spectra = list()
        for hdu in range(1, len(hdulist)):
            if hdulist[hdu].name in ["SPALL", "ZALL", "ZLINE"]:
                continue
            spectra.append(_load_BOSS_HDU(hdulist, hdu, **kwargs))
        return SpectrumList(spectra)


def _load_BOSS_HDU(hdulist: HDUList, hdu: int, **kwargs):
    """
    HDU processor for BOSS spectra redux HDU's

    Parameters
    ----------
    hdulist: HDUList
        HDUList generated from imported file.
    hdu : int
        Specified HDU to load.

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file

    """
    # Fetch flux, e_flux, spectral axis, and units
    # NOTE: flux unit info is not in BOSS spec files.
    flux_unit = Unit("1e-17 erg / (Angstrom cm2 s)")  # NOTE: hardcoded unit
    spectral_axis = Quantity(10**hdulist[hdu].data["LOGLAM"], unit=Angstrom)

    flux = Quantity(hdulist[hdu].data["FLUX"], unit=flux_unit)
    # no e_flux, so we use inverse of variance
    ivar = InverseVariance(hdulist[hdu].data["IVAR"])

    # Fetch mask -- stored in two ways
    try:
        mask = hdulist[hdu].data["MASK"]
    except KeyError:
        mask = hdulist[hdu].data["OR_MASK"]

    # NOTE: specutils considers 0/False as valid values, simlar to numpy convention
    mask = mask != 0

    # Fetch metadata
    # NOTE: specFull file does not include S/N value, but this gets calculated
    #       for mwmVisit/mwmStar files when they are created.
    meta = dict()
    meta["header"] = hdulist[0].header
    meta["name"] = hdulist[hdu].name

    return Spectrum1D(spectral_axis=spectral_axis,
                      flux=flux,
                      uncertainty=ivar,
                      mask=mask,
                      meta=meta)


# MWM LOADERS
@data_loader(
    "SDSS-V mwm",
    identifier=mwm_identify,
    dtype=Spectrum1D,
    priority=20,
    extensions=["fits"],
)
def load_sdss_mwm_1d(file_obj, hdu: Optional[int] = None, **kwargs):
    """
    Load an unspecified spec file as a Spectrum1D.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList..
    hdu : int
        Specified HDU to load.

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file
    """
    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # Check if file is empty first
        datasums = []
        for i in range(1, len(hdulist)):
            datasums.append(int(hdulist[i].header.get("DATASUM")))
        if (np.array(datasums) == 0).all():
            raise ValueError("Specified file is empty.")

        # TODO: how should we handle this -- multiple things in file, but the user cannot choose.
        if hdu is None:
            for i in range(len(hdulist)):
                if hdulist[i].header.get("DATASUM") != "0":
                    hdu = i
                    break

        return _load_mwmVisit_or_mwmStar_hdu(hdulist, hdu, **kwargs)


@data_loader(
    "SDSS-V mwm multi",
    identifier=mwm_identify,
    dtype=SpectrumList,
    priority=20,
    extensions=["fits"],
)
def load_sdss_mwm_list(file_obj, **kwargs):
    """
    Load an mwmStar spec file as a SpectrumList.

    Parameters
    ----------
    file_obj : str, file-like, or HDUList
        FITS file name, file object, or HDUList.

    Returns
    -------
    SpectrumList
        The spectra contained in the file, where:
            Spectrum1D
                A given spectra of nD flux
            None
                If there are no spectra for that spectrograph/observatory
    """
    spectra = SpectrumList()
    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:
        # Check if file is empty first
        datasums = []
        for hdu in range(1, len(hdulist)):
            datasums.append(int(hdulist[hdu].header.get("DATASUM")))
        if (np.array(datasums) == 0).all():
            raise ValueError("Specified file is empty.")

        # Now load file
        for hdu in range(1, len(hdulist)):
            if hdulist[hdu].header.get("DATASUM") == "0":
                # Skip zero data HDU's
                # TODO: validate if we want this printed warning or not.
                # it might get annoying & fill logs with useless alerts.
                print("WARNING: HDU{} ({}) is empty.".format(
                    hdu, hdulist[hdu].name))
                continue
            spectra.append(_load_mwmVisit_or_mwmStar_hdu(hdulist, hdu))
    return spectra


def _load_mwmVisit_or_mwmStar_hdu(hdulist: HDUList, hdu: int, **kwargs):
    """
    HDU loader subfunction for MWM files

    Parameters
    ----------
    hdulist: HDUList
        HDUList generated from imported file.
    hdu: int
        Specified HDU to load.

    Returns
    -------
    Spectrum1D
        The spectrum with nD flux contained in the HDU.

    """
    if hdulist[hdu].header.get("DATASUM") == "0":
        raise IndexError(
            "Attemped to load an empty HDU specified at HDU{}".format(hdu))

    # Fetch wavelength
    # encoded as WCS for visit, and 'wavelength' for star
    try:
        wavelength = np.array(hdulist[hdu].data["wavelength"])[0]
    except KeyError:
        wavelength = _wcs_log_linear(
            hdulist[hdu].header.get("NPIXELS"),
            hdulist[hdu].header.get("CDELT"),
            hdulist[hdu].header.get("CRVAL"),
        )
    finally:
        if wavelength is None:
            raise ValueError(
                "Couldn't find wavelength data in HDU{}.".format(hdu))
        spectral_axis = Quantity(wavelength, unit=Angstrom)

    # Fetch flux, e_flux
    # NOTE:: flux info is not written into mwm files
    flux_unit = Unit("1e-17 erg / (Angstrom cm2 s)")  # NOTE: hardcoded unit
    flux = Quantity(hdulist[hdu].data["flux"], unit=flux_unit)
    e_flux = InverseVariance(array=hdulist[hdu].data["ivar"])

    # Collect bitmask
    mask = hdulist[hdu].data["pixel_flags"]
    # NOTE: specutils considers 0/False as valid values, simlar to numpy convention
    mask = mask != 0

    # collapse shape if 1D spectra in 2D array
    # NOTE: this fixes a jdaviz handling bug for 2D of shape 1,
    #       it could be that it's expected to be parsed this way.
    if flux.shape[0] == 1:
        flux = flux[0]
        e_flux = e_flux[0]
        mask = mask[0]

    # Create metadata
    meta = dict()
    meta["header"] = hdulist[0].header

    # Add SNR to metadata
    meta["snr"] = np.array(hdulist[hdu].data["snr"])

    # Add identifiers (obj, telescope, mjd, datatype)
    # TODO: need to see what metadata we're interested in for the MWM files.
    meta["telescope"] = hdulist[hdu].data["telescope"]
    meta["instrument"] = hdulist[hdu].header.get("INSTRMNT")
    try:
        meta["obj"] = hdulist[hdu].data["obj"]
    except KeyError:
        pass
    try:
        meta["date"] = hdulist[hdu].data["date_obs"]
        meta["mjd"] = hdulist[hdu].data["mjd"]
        meta["datatype"] = "mwmVisit"
    except KeyError:
        meta["mjd"] = (str(hdulist[hdu].data["min_mjd"][0]) + "->" +
                       str(hdulist[hdu].data["max_mjd"][0]))
        meta["datatype"] = "mwmStar"
    finally:
        meta["name"] = hdulist[hdu].name

    return Spectrum1D(
        spectral_axis=spectral_axis,
        flux=flux,
        uncertainty=e_flux,
        mask=mask,
        meta=meta,
    )



import numpy as np
import os
import re
import urllib
import io
import contextlib

from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.utils.exceptions import AstropyUserWarning
import astropy.units as u
import warnings

from specutils.spectra import Spectrum1D


@contextlib.contextmanager
def read_fileobj_or_hdulist(*args, **kwargs):
    """ Context manager for reading a filename or file object

    Returns
    -------
    hdulist : :class:`~astropy.io.fits.HDUList`
        Provides a generator-iterator representing the open file object handle.
    """
    # Access the fileobj or filename arg
    # Do this so identify functions are useable outside of Spectrum1d.read context
    try:
        fileobj = args[2]
    except IndexError:
        fileobj = args[0]

    if isinstance(fileobj, fits.hdu.hdulist.HDUList):
        if fits.util.fileobj_closed(fileobj):
            hdulist = fits.open(fileobj.name, **kwargs)
        else:
            hdulist = fileobj
    elif isinstance(fileobj, io.BufferedReader):
        hdulist = fits.open(fileobj)
    else:
        hdulist = fits.open(fileobj, **kwargs)

    try:
        yield hdulist

    # Cleanup even after identifier function has thrown an exception: rewind generic file handles.
    finally:
        if not isinstance(fileobj, fits.hdu.hdulist.HDUList):
            try:
                fileobj.seek(0)
            except (AttributeError, io.UnsupportedOperation):
                hdulist.close()


def spectrum_from_column_mapping(table, column_mapping, wcs=None, verbose=False):
    """
    Given a table and a mapping of the table column names to attributes
    on the Spectrum1D object, parse the information into a Spectrum1D.

    Parameters
    ----------
    table : :class:`~astropy.table.Table`
        The table object (e.g. returned from ``Table.read('data_file')``).

    column_mapping : dict
        A dictionary describing the relation between the table columns
        and the arguments of the `Spectrum1D` class, along with unit
        information. The dictionary keys should be the table column names
        while the values should be a two-tuple where the first element is the
        associated `Spectrum1D` keyword argument, and the second element is the
        unit for the file column (or ``None`` to take unit from the table header)::

            column_mapping = {'FLUX': ('flux', 'Jy'),
                              'WAVE': ('spectral_axis', 'um')}

    wcs : :class:`~astropy.wcs.WCS` or :class:`gwcs.WCS`
        WCS object passed to the Spectrum1D initializer.

    verbose : bool
        Print extra info.

    Returns
    -------
    :class:`~specutils.Spectrum1D`
        The spectrum with 'spectral_axis', 'flux' and optionally 'uncertainty'
        as identified by `column_mapping`.
    """
    spec_kwargs = {}

    # Associate columns of the file with the appropriate spectrum1d arguments
    for col_name, (kwarg_name, cm_unit) in column_mapping.items():
        # If the table object couldn't parse any unit information,
        # fallback to the column mapper defined unit
        tab_unit = table[col_name].unit

        if tab_unit and cm_unit is not None:
            # If the table unit is defined, retrieve the quantity array for
            # the column
            kwarg_val = u.Quantity(table[col_name], tab_unit)

            # Attempt to convert the table unit to the user-defined unit.
            if verbose:
                print(f"Attempting auto-convert of table unit '{tab_unit}' to "
                      f"user-provided unit '{cm_unit}'.")

            if not isinstance(cm_unit, u.Unit):
                cm_unit = u.Unit(cm_unit)
            if cm_unit.physical_type in ('length', 'frequency', 'energy'):
                # Spectral axis column information
                kwarg_val = kwarg_val.to(cm_unit, equivalencies=u.spectral())
            elif 'spectral flux' in str(cm_unit.physical_type):
                # Flux/error column information
                kwarg_val = kwarg_val.to(cm_unit, equivalencies=u.spectral_density(1 * u.AA))
        elif tab_unit:
            # The user has provided no unit in the column mapping, so we
            # use the unit as defined in the table object.
            kwarg_val = u.Quantity(table[col_name], tab_unit)
        elif cm_unit is not None:
            # In this case, the user has defined a unit in the column mapping
            # but no unit has been defined in the table object.
            kwarg_val = u.Quantity(table[col_name], cm_unit)
        else:
            # Neither the column mapping nor the table contain unit information.
            # This may be desired e.g. for the mask or bit flag arrays.
            kwarg_val = table[col_name]

        # Transpose > 1D data to row-major format
        if kwarg_val.ndim > 1:
            kwarg_val = kwarg_val.T

        spec_kwargs.setdefault(kwarg_name, kwarg_val)

    # Ensure that the uncertainties are a subclass of NDUncertainty
    if spec_kwargs.get('uncertainty') is not None:
        spec_kwargs['uncertainty'] = StdDevUncertainty(
            spec_kwargs.get('uncertainty'))

    return Spectrum1D(**spec_kwargs, wcs=wcs, meta={'header': table.meta})


def generic_spectrum_from_table(table, wcs=None):
    """
    Load spectrum from an Astropy table into a Spectrum1D object.
    Uses the following logic to figure out which column is which:

     * Spectral axis (dispersion) is the first column with units
     compatible with ``u.spectral()`` or with length units such as 'pix'.
     Need not be present, if a valid ``wcs`` parameter is passed.

     * Flux is taken from the first column with units compatible with
     ``u.spectral_density()``, or with other likely culprits such as
     'adu' or 'cts/s'.

     * Uncertainty comes from the next column with the same units as flux.

    Parameters
    ----------
    table : :class:`~astropy.table.Table`
        Table containing a column of ``flux``, and optionally ``spectral_axis``
        and ``uncertainty`` as defined above.
    wcs : :class:`~astropy.wcs.WCS`
        A FITS WCS object. If this is present, the machinery will fall back
        and default to using the ``wcs`` to find the dispersion information.

    Returns
    -------
    :class:`~specutils.Spectrum1D`
        The spectrum that is represented by the data from the columns
        as automatically identified above.

    Raises
    ------
    Warns if uncertainty has zeros or negative numbers.
    Raises IOError if it can't figure out the columns.

    """
    # Local function to find the wavelength or frequency column
    def _find_spectral_axis_column(table, columns_to_search):
        """
        Figure out which column in a table holds the spectral axis (dispersion).
        Take the first column that has units compatible with u.spectral()
        equivalencies. If none meet that criterion, look for other likely
        length units such as 'pix'.
        """
        additional_valid_units = [u.Unit('pix')]
        found_column = None

        # First, search for a column with units compatible with Angstroms
        for c in columns_to_search:
            try:
                table[c].to("AA", equivalencies=u.spectral())
                found_column = c
                break
            except Exception:
                continue

        # If no success there, check for other possible length units
        if found_column is None:
            for c in columns_to_search:
                if table[c].unit in additional_valid_units:
                    found_column = c
                    break

        return found_column

    # Local function to find the flux column
    def _find_spectral_column(table, columns_to_search, spectral_axis):
        """
        Figure out which column in a table holds the fluxes or uncertainties.
        Take the first column that has units compatible with
        u.spectral_density() equivalencies. If none meet that criterion,
        look for other likely length units such as 'adu' or 'cts/s'.
        """
        additional_valid_units = [u.Unit('adu'), u.Unit('ct/s'), u.Unit('count')]
        found_column = None

        # First, search for a column with units compatible with Jansky
        for c in columns_to_search:
            try:
                # Check for multi-D flux columns
                if table[c].ndim == 1:
                    spec_ax = spectral_axis
                else:
                    # Assume leading dimension corresponds to spectral_axis
                    spec_shape = np.ones(table[c].ndim, dtype=int)
                    spec_shape[0] = -1
                    spec_ax = spectral_axis.reshape(spec_shape)
                table[c].to("Jy", equivalencies=u.spectral_density(spec_ax))
                found_column = c
                break
            except Exception:
                continue

        # If no success there, check for other possible flux units
        if found_column is None:
            for c in columns_to_search:
                if table[c].unit in additional_valid_units:
                    found_column = c
                    break

        return found_column

    # Make a copy of the column names so we can remove them as they are found
    colnames = table.colnames.copy()

    # Use the first column that has spectral unit as the dispersion axis
    spectral_axis_column = _find_spectral_axis_column(table, colnames)

    if spectral_axis_column is None and wcs is None:
        raise IOError("Could not identify column containing the wavelength, frequency or energy")
    elif wcs is not None:
        spectral_axis = None
    else:
        spectral_axis = table[spectral_axis_column].to(table[spectral_axis_column].unit)
        colnames.remove(spectral_axis_column)

    # Use the first column that has a spectral_density equivalence as the flux
    flux_column = _find_spectral_column(table, colnames, spectral_axis)
    if flux_column is None:
        raise IOError("Could not identify column containing the flux")
    flux = table[flux_column].to(table[flux_column].unit)
    colnames.remove(flux_column)
    # For > 1D data transpose to row-major format
    if flux.ndim > 1:
        flux = flux.T

    # Use the next column with the same units as flux as the uncertainty
    # Interpret it as a standard deviation and check if it has zeros or negative values
    err_column = None
    for c in colnames:
        if table[c].unit == table[flux_column].unit:
            err_column = c
            break
    if err_column is not None:
        if table[err_column].ndim > 1:
            err = table[err_column].T
        elif flux.ndim > 1:  # Repeat uncertainties over all flux columns
            err = np.tile(table[err_column], flux.shape[0], 1)
        else:
            err = table[err_column]
        err = StdDevUncertainty(err.to(err.unit))
        if np.min(table[err_column]) <= 0.:
            warnings.warn("Standard Deviation has values of 0 or less", AstropyUserWarning)
    else:
        err = None

    # Check for mask
    if 'mask' in table.colnames:
        mask = table['mask']
        if mask.ndim > 1:
            mask = mask.T
    else:
        mask = None

    # Create the Spectrum1D object and return it
    if wcs is not None or spectral_axis_column is not None and flux_column is not None:
        # For > 1D spectral axis transpose to row-major format and return SpectrumCollection
        spectrum = Spectrum1D(flux=flux, spectral_axis=spectral_axis,
                              uncertainty=err, meta={'header': table.meta}, wcs=wcs,
                              mask=mask)

    return spectrum


def _fits_identify_by_name(origin, fileinp, *args,
                           pattern=r'(?i).*\.fit[s]?$', **kwargs):
    """
    Check whether input file is FITS and matches a given name pattern.
    Utility function to construct an `identifier` for Astropy I/O Registry.

    Parameters
    ----------
    fileinp : str or file-like object
        FITS file name or object (provided from name by Astropy I/O Registry).
    pattern : regex str or re.Pattern
        File name pattern to be matched.
        Note: loaders should define a pattern sufficiently specific for their
        spectrum file types to avoid ambiguous/multiple matches.
    """
    fileobj = None
    filepath = None
    if pattern is None:
        pattern = r''
    _spec_pattern = re.compile(pattern)

    if isinstance(fileinp, str):
        filepath = fileinp
        try:
            fileobj = open(filepath, mode='rb')
        except FileNotFoundError:
            # Check if path points to valid url
            try:
                fileinp = urllib.request.urlopen(filepath)
            except ValueError:
                return False
    elif fits.util.isfile(fileinp):
        fileobj = fileinp
        filepath = fileobj.name

    # Check for `urlopen` object - can only probe content if seekable
    if hasattr(fileinp, 'url') and hasattr(fileinp, 'seekable'):
        filepath = urllib.parse.unquote(fileinp.url)
        if fileinp.seekable():
            fileobj = fileinp

    check = (_spec_pattern.match(os.path.basename(filepath)) is not None and
             fits.connect.is_fits(origin, filepath, fileobj, *args))

    if fileobj is not None:
        fileobj.close()

    return check



import os
import glob
import warnings
from itertools import chain

import astropy.units as u
from astropy.units import Quantity
from astropy.table import Table
from astropy.io import fits
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, InverseVariance
from astropy.time import Time
from astropy.wcs import WCS
from gwcs.wcstools import grid_from_bounding_box

from ...spectra import Spectrum1D, SpectrumList
from ..registers import data_loader
from ..parsing_utils import read_fileobj_or_hdulist

__all__ = ["jwst_x1d_single_loader", "jwst_x1d_multi_loader", "jwst_x1d_miri_mrs_loader"]


def identify_jwst_miri_mrs(origin, *args, **kwargs):
    """
    Check whether the given set of files is a JWST MIRI MRS spectral data product.
    """
    input = args[2]

    # if string, it can be either a directory or a glob pattern (this last
    # one not implemented yet due to astropy choking when passed an invalid
    # file path string).
    if isinstance(input, str):
        if os.path.isdir(input):
            return True

        if len(glob.glob(input)) > 0:
            return True

    # or it can be either a list of file names, or a list of file objects
    elif isinstance(input, (list, tuple)) and len(input) > 0:
        return True

    return False


def _identify_spec1d_fits(origin, extname, *args, **kwargs):
    """ Generic spec 1d identifier function """
    is_jwst = _identify_jwst_fits(*args)
    with read_fileobj_or_hdulist(*args, memmap=False, **kwargs) as hdulist:
        return (is_jwst and extname in hdulist and (extname, 2) not in hdulist)


def _identify_spec1d_multi_fits(origin, extname, *args, **kwargs):
    """
    Check whether the given file is a JWST c1d/x1d spectral data product with many slits.
    """
    is_jwst = _identify_jwst_fits(*args)
    with read_fileobj_or_hdulist(*args, memmap=False, **kwargs) as hdulist:
        return is_jwst and (extname, 2) in hdulist


def identify_jwst_c1d_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST c1d spectral data product.
    """
    return _identify_spec1d_fits(origin, 'COMBINE1D', *args, **kwargs)


def identify_jwst_c1d_multi_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST c1d spectral data product with many slits.
    """
    return _identify_spec1d_multi_fits(origin, 'COMBINE1D', *args, **kwargs)


def identify_jwst_x1d_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST x1d spectral data product.
    """
    return _identify_spec1d_fits(origin, 'EXTRACT1D', *args, **kwargs)


def identify_jwst_x1d_multi_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST x1d spectral data product with many slits.
    """
    return _identify_spec1d_multi_fits(origin, 'EXTRACT1D', *args, **kwargs)


def identify_jwst_s2d_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST s2d spectral data product.
    """
    is_jwst = _identify_jwst_fits(*args)
    with read_fileobj_or_hdulist(*args, memmap=False, **kwargs) as hdulist:
        return (is_jwst and "SCI" in hdulist and ("SCI", 2) not in hdulist
                and "EXTRACT1D" not in hdulist and len(hdulist["SCI"].data.shape) == 2)


def identify_jwst_s2d_multi_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST s2d spectral data product with many slits.
    """
    is_jwst = _identify_jwst_fits(*args)
    with read_fileobj_or_hdulist(*args, memmap=False, **kwargs) as hdulist:
        return (is_jwst and ("SCI", 2) in hdulist and "EXTRACT1D" not in hdulist
                and len(hdulist["SCI"].data.shape) == 2)


def identify_jwst_s3d_fits(origin, *args, **kwargs):
    """
    Check whether the given file is a JWST s3d spectral data product.
    """
    is_jwst = _identify_jwst_fits(*args)
    with read_fileobj_or_hdulist(*args, memmap=False, **kwargs) as hdulist:
        return (is_jwst and "SCI" in hdulist and "EXTRACT1D" not in hdulist
                and len(hdulist["SCI"].data.shape) == 3)


def _identify_jwst_fits(*args):
    """
    Check whether the given file is a JWST data product.
    """
    try:
        with read_fileobj_or_hdulist(*args, memmap=False) as hdulist:
            return "ASDF" in hdulist and hdulist[0].header.get("TELESCOP") == "JWST"
    # This probably means we didn't have a FITS file
    except Exception:
        return False


@data_loader(
    "JWST c1d", identifier=identify_jwst_c1d_fits, dtype=Spectrum1D,
    extensions=['fits'], priority=10,
)
def jwst_c1d_single_loader(file_obj, **kwargs):
    """
    Loader for JWST c1d 1-D spectral data in FITS format

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file.
    """
    spectrum_list = _jwst_spec1d_loader(file_obj, extname='COMBINE1D', **kwargs)
    if len(spectrum_list) == 1:
        return spectrum_list[0]
    else:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra. "
                           "Use SpectrumList.read() instead.")


@data_loader(
    "JWST c1d multi", identifier=identify_jwst_c1d_multi_fits,
    dtype=SpectrumList, extensions=['fits'], priority=10,
)
def jwst_c1d_multi_loader(file_obj, **kwargs):
    """
    Loader for JWST x1d 1-D spectral data in FITS format

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    SpectrumList
        A list of the spectra that are contained in the file.
    """
    return _jwst_spec1d_loader(file_obj, extname='COMBINE1D', **kwargs)


@data_loader(
    "JWST x1d", identifier=identify_jwst_x1d_fits, dtype=Spectrum1D,
    extensions=['fits'], priority=10
)
def jwst_x1d_single_loader(file_obj, **kwargs):
    """
    Loader for JWST x1d 1-D spectral data in FITS format

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file.
    """
    spectrum_list = _jwst_spec1d_loader(file_obj, extname='EXTRACT1D', **kwargs)
    if len(spectrum_list) == 1:
        return spectrum_list[0]
    else:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra. "
                           "Use SpectrumList.read() instead.")


@data_loader(
    "JWST x1d multi", identifier=identify_jwst_x1d_multi_fits,
    dtype=SpectrumList, extensions=['fits'], priority=10,
)
def jwst_x1d_multi_loader(file_obj, **kwargs):
    """
    Loader for JWST x1d 1-D spectral data in FITS format

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    SpectrumList
        A list of the spectra that are contained in the file.
    """
    return _jwst_spec1d_loader(file_obj, extname='EXTRACT1D', **kwargs)


@data_loader(
    "JWST x1d MIRI MRS", identifier=identify_jwst_miri_mrs, dtype=SpectrumList,
    extensions=['*'], priority=10,
)
def jwst_x1d_miri_mrs_loader(input, missing="raise", **kwargs):
    """
    Loader for JWST x1d MIRI MRS spectral data in FITS format.

    A single data set consists of a bunch of _x1d files corresponding to
    a variety of wavelength bands. This reader reads them one by one and packs
    the result into a SpectrumList instance.

    Parameters
    ----------
    input : list of str or file-like
        List of FITS file names, or objects (provided from name by
        Astropy I/O Registry). Alternatively, a directory path on
        which glob.glob runs with pattern an implicit pattern "_x1d.fits",
        or a directory path with a glob pattern already set.
    missing : {'warn', 'silent'}
        Allows the user to continue loading if one file is missing by setting
        the value to "warn" or "silent". In the first case a warning will be issued
        to the user, in the latter the file will silently be skipped. Any other
        value will result in a FileNotFoundError if any files in the list are missing.

    Returns
    -------
    SpectrumList
        A list of the spectra that are contained in all the files.
    """

    # If input is a list, go read each file. If directory, glob-expand
    # list of file names.
    if not isinstance(input, (list, tuple)):
        if os.path.isdir(input):
            file_list = glob.glob(os.path.join(input, "*_x1d.fits"), recursive=True)
        else:
            file_list = glob.glob(input, recursive=True)
    else:
        file_list = input

    spectra = []
    for file_obj in file_list:
        try:
            sp = _jwst_spec1d_loader(file_obj, **kwargs)
        except FileNotFoundError as e:
            if missing.lower() == "warn":
                warnings.warn(f'Failed to load {file_obj}: {repr(e)}')
                continue
            elif missing.lower() == "silent":
                continue
            else:
                raise FileNotFoundError(f"Failed to load {file_obj}: {repr(e)}. "
                                        "To suppress this error, set argument missing='warn'")

        spectra.append(sp)

    # the call to `chain.from_iterable` allows us to handle multiple HDUs
    # stored within multiple FITS files, all unpacked into one `SpectrumList`
    return SpectrumList(chain.from_iterable(spectra))


def _jwst_spec1d_loader(file_obj, extname='EXTRACT1D', **kwargs):
    """Implementation of loader for JWST x1d 1-D spectral data in FITS format

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).
    extname: str
        The name of the science extension. Either "COMBINE1D" or "EXTRACT1D".  By default "EXTRACT1D".

    Returns
    -------
    SpectrumList
        A list of the spectra that are contained in the file.
    """

    if extname not in ['COMBINE1D', 'EXTRACT1D']:
        raise ValueError('Incorrect extname given for 1d spectral data.')

    spectra = []

    with read_fileobj_or_hdulist(file_obj, memmap=False, **kwargs) as hdulist:

        primary_header = hdulist["PRIMARY"].header

        for hdu in hdulist:
            # Read only the BinaryTableHDUs named COMBINE1D/EXTRACT1D and SCI
            if hdu.name != extname:
                continue

            # Correct some known bad unit strings before reading the table
            bad_units = {"(MJy/sr)^2": "MJy2 sr-2"}
            for c in hdu.columns:
                if c.unit in bad_units:
                    c.unit = bad_units[c.unit]

            data = Table.read(hdu)

            wavelength = Quantity(data["WAVELENGTH"])

            # Determine if FLUX or SURF_BRIGHT column should be returned
            # based on whether it is point or extended source.
            #
            # SRCTYPE used to be in primary header, but was moved around. As
            # per most recent pipeline definition, it should be in the
            # EXTRACT1D extension.
            #
            # SRCTYPE should either be POINT or EXTENDED.  In some cases, it is UNKNOWN
            # or missing.  If that's the case, default to using POINT as the SRCTYPE.
            # Error out only when SRCTYPE is a bad value.
            srctype = None
            if "srctype" in hdu.header:
                srctype = hdu.header.get("srctype", None)

            # checking if SRCTPYE is missing or UNKNOWN
            if not srctype or srctype == 'UNKNOWN':
                warnings.warn('SRCTYPE is missing or UNKNOWN in JWST x1d loader. '
                              'Defaulting to srctype="POINT".')
                srctype = 'POINT'

            if srctype == "POINT":
                flux = Quantity(data["FLUX"])
                if 'ERROR' in data.colnames:
                    uncertainty = StdDevUncertainty(data["ERROR"])
                elif 'FLUX_ERROR' in data.colnames:
                    uncertainty = StdDevUncertainty(data["FLUX_ERROR"])
                else:
                    uncertainty = None
            elif srctype == "EXTENDED":
                flux = Quantity(data["SURF_BRIGHT"])
                uncertainty = StdDevUncertainty(data["SB_ERROR"])
            else:
                raise RuntimeError(f"Keyword SRCTYPE is {srctype}.  It should "
                                   "be 'POINT' or 'EXTENDED'. Can't decide between `flux` and "
                                   "`surf_bright` columns.")

            # Merge primary and slit headers and dump into meta
            slit_header = hdu.header
            header = primary_header.copy()
            header.extend(slit_header, strip=True, update=True)
            meta = {'header': header}

            spec = Spectrum1D(flux=flux, spectral_axis=wavelength,
                              uncertainty=uncertainty, meta=meta)
            spectra.append(spec)

    return SpectrumList(spectra)


@data_loader(
    "JWST s2d", identifier=identify_jwst_s2d_fits, dtype=Spectrum1D,
    extensions=['fits'], priority=10,
)
def jwst_s2d_single_loader(filename, **kwargs):
    """
    Loader for JWST s2d 2D rectified spectral data in FITS format.

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file.
    """
    spectrum_list = _jwst_s2d_loader(filename, **kwargs)
    if len(spectrum_list) == 1:
        return spectrum_list[0]
    elif len(spectrum_list) > 1:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra. "
                           "Use SpectrumList.read() instead.")
    else:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra.")


@data_loader(
    "JWST s2d multi", identifier=identify_jwst_s2d_multi_fits,
    dtype=SpectrumList, extensions=['fits'], priority=10,
)
def jwst_s2d_multi_loader(filename, **kwargs):
    """
    Loader for JWST s2d 2D rectified spectral data in FITS format.

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    SpectrumList
        The spectra contained in the file.
    """
    return _jwst_s2d_loader(filename, **kwargs)


def _jwst_s2d_loader(filename, **kwargs):
    """
    Loader for JWST s2d 2D rectified spectral data in FITS format.

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    SpectrumList
        The spectra contained in the file.
    """
    from stdatamodels import asdf_in_fits

    spectra = []
    slits = None

    # Get a list of GWCS objects from the slits
    with fits.open(filename, memmap=False) as hdulist, asdf_in_fits.open(hdulist) as af:
        # Slits can be listed under "slits", "products" or "exposures"
        if "products" in af.tree:
            slits = "products"
        elif "exposures" in af.tree:
            slits = "exposures"
        elif "slits" in af.tree:
            slits = "slits"
        else:
            # No slits.  Case for s3d cubes
            wcslist = [af.tree["meta"]["wcs"]]

        # Create list of the GWCS objects, one for each slit
        if slits is not None:
            wcslist = [slit["meta"]["wcs"] for slit in af.tree[slits]]

        primary_header = hdulist["PRIMARY"].header

        hdulist_sci = [hdu for hdu in hdulist if hdu.name == "SCI"]

        for hdu, wcs in zip(hdulist_sci, wcslist):
            # Get flux
            try:
                flux_unit = u.Unit(hdu.header["BUNIT"])
            except (ValueError, KeyError):
                flux_unit = None

            # The dispersion axis is 1 or 2.  1=x, 2=y.
            dispaxis = hdu.header.get("DISPAXIS")
            if dispaxis is None:
                dispaxis = hdulist["PRIMARY"].header.get("DISPAXIS")

            # Get the wavelength array from the GWCS object which returns a
            # tuple of (RA, Dec, lambda)
            grid = grid_from_bounding_box(wcs.bounding_box)
            _, _, lam = wcs(*grid)
            _, _, lam_unit = wcs.output_frame.unit

            # Make sure the dispersion axis is the same for every spatial axis
            # for s2d data
            if dispaxis == 1:
                flux_array = hdu.data
                wavelength_array = lam[0]
                # Make sure all rows are the same
                if not (lam == wavelength_array).all():
                    raise RuntimeError("This 2D or 3D spectrum is not rectified "
                                       "and cannot be loaded into a Spectrum1D object.")
            elif dispaxis == 2:
                flux_array = hdu.data.T
                wavelength_array = lam[:, 0]
                # Make sure all columns are the same
                if not (lam.T == lam[None, :, 0]).all():
                    raise RuntimeError("This 2D or 3D spectrum is not rectified "
                                       "and cannot be loaded into a Spectrum1D object.")
            else:
                raise RuntimeError("This 2D spectrum has an unknown dispaxis "
                                   "and cannot be loaded into a Spectrum1D object.")

            flux = Quantity(flux_array, unit=flux_unit)
            wavelength = Quantity(wavelength_array, unit=lam_unit)

            # Merge primary and slit headers and dump into meta
            slit_header = hdu.header
            header = primary_header.copy()
            header.extend(slit_header, strip=True, update=True)
            meta = {'header': header}

            spec = Spectrum1D(flux=flux, spectral_axis=wavelength, meta=meta)
            spectra.append(spec)

    return SpectrumList(spectra)


@data_loader(
    "JWST s3d", identifier=identify_jwst_s3d_fits, dtype=Spectrum1D,
    extensions=['fits'], priority=10,
)
def jwst_s3d_single_loader(filename, **kwargs):
    """
    Loader for JWST s3d 3D rectified spectral data in FITS format.

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    Spectrum1D
        The spectrum contained in the file.
    """
    spectrum_list = _jwst_s3d_loader(filename, **kwargs)
    if len(spectrum_list) == 1:
        return spectrum_list[0]
    elif len(spectrum_list) > 1:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra. "
                           "Use SpectrumList.read() instead.")
    else:
        raise RuntimeError(f"Input data has {len(spectrum_list)} spectra.")


def _jwst_s3d_loader(filename, **kwargs):
    """
    Loader for JWST s3d 3D rectified spectral data in FITS format.

    Parameters
    ----------
    filename : str
        The path to the FITS file

    Returns
    -------
    SpectrumList
        The spectra contained in the file.
    """
    from stdatamodels import asdf_in_fits

    spectra = []

    # Get a list of GWCS objects from the slits
    with fits.open(filename, memmap=False) as hdulist, asdf_in_fits.open(hdulist) as af:
        wcslist = [af.tree["meta"]["wcs"]]

        primary_header = hdulist["PRIMARY"].header

        hdulist_sci = [hdu for hdu in hdulist if hdu.name == "SCI"]

        for hdu, wcs in zip(hdulist_sci, wcslist):
            # Get flux
            try:
                flux_unit = u.Unit(hdu.header["BUNIT"])
            except (ValueError, KeyError):
                flux_unit = None

            # The spectral axis is first.  We need it last
            flux_array = hdu.data.T
            flux = Quantity(flux_array, unit=flux_unit)

            # Get the wavelength array from the GWCS object which returns a
            # tuple of (RA, Dec, lambda).
            # Since the spatial and spectral axes are orthogonal in s3d data,
            # it is much faster to compute a slice down the spectral axis.
            grid = grid_from_bounding_box(wcs.bounding_box)[:, :, 0, 0]
            _, _, wavelength_array = wcs(*grid)
            _, _, wavelength_unit = wcs.output_frame.unit

            wavelength = Quantity(wavelength_array, unit=wavelength_unit)

            # The GWCS is currently broken for some IFUs, here we work around that
            wcs = None
            if wavelength.shape[0] != flux.shape[-1]:
                # Need MJD-OBS for this workaround
                if 'MJD-OBS' not in hdu.header:
                    for key in ('MJD-BEG', 'DATE-OBS'):  # Possible alternatives
                        if key in hdu.header:
                            if key.startswith('MJD'):
                                hdu.header['MJD-OBS'] = hdu.header[key]
                                break
                            else:
                                t = Time(hdu.header[key])
                                hdu.header['MJD-OBS'] = t.mjd
                                break
                wcs = WCS(hdu.header)
                # Swap to match the flux transpose
                wcs = wcs.swapaxes(-1, 0)

            # Merge primary and slit headers and dump into meta
            slit_header = hdu.header
            header = primary_header.copy()
            header.extend(slit_header, strip=True, update=True)
            meta = {'header': header}

            # get uncertainty information
            ext_name = primary_header.get("ERREXT", "ERR")
            err_type = hdulist[ext_name].header.get("ERRTYPE", 'ERR')
            err_unit = hdulist[ext_name].header.get("BUNIT", None)
            err_array = hdulist[ext_name].data.T

            # ERRTYPE can be one of "ERR", "IERR", "VAR", "IVAR"
            # but mostly ERR for JWST cubes
            # see https://jwst-pipeline.readthedocs.io/en/latest/jwst/data_products/science_products.html#s3d
            if err_type == "ERR":
                err = StdDevUncertainty(err_array, unit=err_unit)
            elif err_type == 'VAR':
                err = VarianceUncertainty(err_array, unit=err_unit)
            elif err_type == 'IVAR':
                err = InverseVariance(err_array, unit=err_unit)
            elif err_type == 'IERR':
                warnings.warn("Inverse error is not yet a supported astropy.nddata "
                              "uncertainty. Setting err to None.")
                err = None

            # get mask information
            mask_name = primary_header.get("MASKEXT", "DQ")
            mask = hdulist[mask_name].data.T

            if wcs is not None:
                spec = Spectrum1D(flux=flux, wcs=wcs, meta=meta, uncertainty=err, mask=mask)
            else:
                spec = Spectrum1D(flux=flux, spectral_axis=wavelength, meta=meta,
                                  uncertainty=err, mask=mask)
            spectra.append(spec)

    return SpectrumList(spectra)
